package exampler.fitlibrary;

import fitlibrary.FunctionFixture;
import fitlibrary.parse.Row;
import fitlibrary.parse.Table;

import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;

public abstract class BooleanFilterFixture extends FunctionFixture {
    static int ORIG_NUM_HEADING_ROWS = 2;
    static int ORIG_TABLE_DESCRIPTION_ROW = 0;
    static int ORIG_TABLE_TYPE_CELL = 0;
    static int ORIG_DERIVED_FIXTURE_CELL = 1;
    static int ORIG_CATEGORY_COLUMN = 0;
    static final int ORIG_EXAMPLE_COLUMN = 1;
    static final int ORIG_COUNTEREXAMPLE_COLUMN = 2;

    // which is flipper and which is maintainer is determined by subclass.
    // TODO: look for column headings instead of assume which column is which.
    protected Table originalTable;
    protected String derivedFixtureName;
    protected String[] derivedTableHeadings;
    protected Map /* <String, RepeatingStream> */ flippers;
    protected Map /* <String, RepeatingStream> */ maintainers;
    protected int flipperColumn = -1;




    // Inherited methods - the root of the work.

    protected void processRow(Row row) {
        // Does nothing - all the work's in teardown.
    }

    protected void bindFirstRowToTarget(Row row) {
        // Does nothing - all the work's in teardown.
    }

    protected void doTearDown(Table originalTable) {
        useTable(originalTable);
        String[][] flipperRows = flipperRows();
        String[][] maintainerRows = maintainerRows();
        TableParse derivedTable = assemble(derivedFixtureName, derivedTableHeadings, flipperRows, maintainerRows);
        derivedTable.insertAfter(originalTable);
    }



    // What subclasses use to make themselves special

    protected void useTable(Table originalTable) {
        this.originalTable = originalTable;
        derivedFixtureName = derivedFixture();
        derivedTableHeadings = derivedHeadings();
        setFlippersAndMaintainers();
    }


    protected abstract void setFlippersAndMaintainers();  // subclasses call version with args.

    protected void setFlippersAndMaintainers(int flipperColumn, int maintainerColumn) {
        this.flipperColumn = flipperColumn;
        this.flippers = columnToMap(flipperColumn);
        this.maintainers = columnToMap(maintainerColumn);
    }

    protected abstract String resultOfFlipping();
    protected abstract String resultOfMaintaining();


    // Routines that select pieces from the original
    // table handed to this fixture.
    // The "derived" values are those used to construct
    // the resulting table that does the actual test.


    protected String derivedFixture() {
        return originalTable.row(ORIG_TABLE_DESCRIPTION_ROW).cell(ORIG_DERIVED_FIXTURE_CELL).text();
    }

    protected String[] derivedHeadings() {
        return columnToArray(ORIG_CATEGORY_COLUMN);
    }

    private int numCategories() {
        String[] categories = derivedHeadings();
        return categories.length;
    }

    private String[] cellTokens(String containingTokens) {
        return containingTokens.split("\\s*,\\s*");
    }

    private int countTokensInColumn(int column) {
        int result = 0;
        String[] allEntries = columnToArray(column);
        for (int i = 0; i < allEntries.length; i++) {
            String oneEntry = allEntries[i];
            result += cellTokens(oneEntry).length;
        }
        return result;
    }

    protected String columnDefault(int column) {
        return column == ORIG_EXAMPLE_COLUMN ? "true" : "false";
    }
    private String[] columnToArray(int column) {
        ArrayList /* <String> */ result = new ArrayList/*<String>*/();
        for(int i = ORIG_NUM_HEADING_ROWS; i < originalTable.size(); i++) {
            Row row = originalTable.row(i);
            if (row.size() <= column) {
                result.add(columnDefault(column));
            } else {
                result.add(row.cell(column).text());
            }
        }
        return (String[]) result.toArray(new String[result.size()]);
    }

    protected Map /* <String, RepeatingStream> */ columnToMap(int columnIndex) {
        String[] column = columnToArray(columnIndex);
        HashMap /*<String,RepeatingStream>*/ result = new HashMap/*<String, RepeatingStream>*/();
        for (int i = 0; i < column.length; i++) {
            // TODO: this should use more sophisticated decomposition.
            String[] decomposed = cellTokens(column[i]);
            result.put(derivedTableHeadings[i], new RepeatingStream(decomposed));
        }
        return result;
    }

    private void startValuesAtBeginning(Map/*<String, RepeatingStream>*/ map) {
        for (java.util.Iterator it = map.values().iterator(); it.hasNext();) {
            RepeatingStream stream = (RepeatingStream) it.next();
            stream.reset();
        }
    }


    // The Flipper rows are those that show a particular variable
    // can, in itself, affect the truth value of an entire expression.

    public int flipperRowCount() {
        return countTokensInColumn(flipperColumn);
    }

    protected String[][] flipperRows() {
        int numTestRows = flipperRowCount();

        String[][] result = new String[numTestRows][numCategories()];
        placeFlippingValues(result);
        fillBlanksWithMaintainers(result);
        // Could fill all the blanks first, but then
        // the maintainer values in a column would
        // appear to skip values. That'd be harmless since
        // all values will be used in the maintainer rows,
        // but it might alarm the user.

        return result;
    }

    private void placeFlippingValues(String[][] result) {
        int cellindex = -1;
        int remainingFlippers = 0;
        RepeatingStream currentStream = null;
        startValuesAtBeginning(flippers);
        for (int i = 0; i < result.length; i++) {
            String[] row = result[i];
            if (remainingFlippers == 0) {
                cellindex++;
                currentStream = (RepeatingStream) flippers.get(derivedTableHeadings[cellindex]);
                remainingFlippers = currentStream.cycle();
            }
            row[cellindex] = highlighted(currentStream.next());
            remainingFlippers--;
        }
    }

    private String highlighted(String s) {
        return "<span style=\"font-size: 1.3em; font-style: italic; font-weight: bold;\">" + s + "</span>";
    }

    private void fillBlanksWithMaintainers(String[][] result) {
        startValuesAtBeginning(maintainers);
        for (int i = 0; i < result.length; i++) {
            String[] row = result[i];
            for (int cellindex = 0; cellindex < result[0].length; cellindex++) {
                if (row[cellindex] == null) {
                    RepeatingStream stream = (RepeatingStream) maintainers.get(derivedTableHeadings[cellindex]);
                    row[cellindex] = stream.next();
                }
            }
        }
    }


    // The maintainer rows show that no maintainer value flips
    // the truth value. The flipper rows don't show that because
    // they don't rule out a maintainer value that flips (because
    // the intentional flipper value would mask that).


    public int maintainerRowCount() {
        startValuesAtBeginning(maintainers);
        int retval = 0;
        for (int i = 0; i < numCategories(); i++) {
            RepeatingStream stream = (RepeatingStream) maintainers.get(derivedTableHeadings[i]);
            int numMaintainers = stream.baseArray().length;
            retval = Math.max(retval, numMaintainers);
        }
        return retval;
    }

    protected String[][] maintainerRows() {
        int numTestRows = maintainerRowCount();

        String[][] result = new String[numTestRows][numCategories()];

        startValuesAtBeginning(maintainers);
        for (int i = 0; i < result.length; i++) {
            String[] row = result[i];
            for (int cellindex = 0; cellindex < row.length; cellindex++) {
                RepeatingStream stream = (RepeatingStream) maintainers.get(derivedTableHeadings[cellindex]);
                row[cellindex] = stream.next();
            }
        }
        return result;
    }



    // Once the parts are calculated, they are assembled into
    // a new Fit table.

    protected TableParse assemble(String derivedFixtureName, String[] headings, String[][] flipperRows, String[][] maintainerRows) {
        TableParse retval = new TableParse();

        assembleHeader(retval, derivedFixtureName, headings);
        addTestRows(retval, flipperRows, highlighted(resultOfFlipping()));
        addTestRows(retval, maintainerRows, resultOfMaintaining());
        annotate(retval);
        return retval;
    }

    private void assembleHeader(TableParse retval, String derivedFixtureName, String[] headings) {
        RowParse fixtureRow = new RowParse();
        RowParse headingRow = new RowParse();
        retval.addAsLastRow(fixtureRow);
        retval.addAsLastRow(headingRow);

        fixtureRow.addAsLastCell(derivedFixtureName);

        for (int i = 0; i < headings.length; i++) {
            String heading = headings[i];
            headingRow.addAsLastCell(heading);
        }
        headingRow.addAsLastCell("result()");
    }

    private void addTestRows(TableParse retval, String[][] rows, String resultValue) {
        for (int i = 0; i < rows.length; i++) {
            String[] row = rows[i];
            RowParse resultRow = RowParse.fromStrings(row);
            resultRow.addAsLastCell(resultValue);
            retval.addAsLastRow(resultRow);
        }
    }

    private void annotate(TableParse derivedTable) {
        derivedTable.leader = "<blockquote><p>The following table is an expansion of the previous one.</p></blockquote>  ";
    }



}
